/* main.c
   
   See the README file for usage instructions, and INSTALL
   for help on installing
   
*/


/*
  SquidClamav - A Clamav Antivirus Redirector for Squid

  This project is a modified version of the excellent Squirm Redirector
  for Squid Maintained by Chris Foote, and copyrighted
  as follow :

  Copyright (C) 1998 Chris Foote & Wayne Piekarski

  The original Squirm version used was squirm-1.0betaB. Some other parts
  are cut and paste from the ex1.c program given in the ClamAv distribution
  and are copyrighted: Copyright (C) 2002 - 2004 Tomasz Kojm

  All other code: Copyright (C) 2005 Gilles Darold - Groupe SAMSE

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
    
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  Please see the file GPL in this directory for full copyright
  information.
*/

#define TRUE  1
#define FALSE 0
#define FSIZE 102400         /* scan max 100kb */

#include "paths.h"
#include "squidclamav.h"
#include "log.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/types.h>
#include <ctype.h>
#include "lists.h"
#include <clamav.h>

#include <curl/curl.h>
#include <signal.h>

void sighup(); /* routine to reload virus database on SIGHUP signal */
void sigterm(); /* routine to quit with cleaning on SIGTERM signal (kill -15) */

int bridge_mode = 0;
int interactive = 0;
char *alternate_config = NULL;
char *redirect_url = NULL;
int maxsize_url = FSIZE;
int check_content_type = 0;
int has_regex = 0;
int size_limit = 0;

struct cl_node *root = NULL;
FILE *tf;

/*  exit values:
    0 = normal
    1 = file errors
    2 = abnormal errors
    3 = malloc problems
    
    Under normal conditions, SquidClamav will never abort (you don't
    want your Squid to suddenly die if you've put an invalid
    regex into your config file, or renamed the config file!  
    Under these abnormal conditions, squidclamav will run in bridge_mode 
    where stdin is echoed to stdout!

    If you've entered an invalid regex expression that regcomp()
    doesn't like, or perhaps an invalid network number, check it 
    out in the ERROR_LOG.
    
    It should never ever abort under normal circumstances - even
    if it can't allocate memory (which shouldn't happen, and Squid 
    would be running like a dog under low memory conditions anyway) 
    bridge_mode will be set and it will keep running :-)
    
*/

int main(int argc, char **argv)
{
  char buff[MAX_BUFF];
  struct IN_BUFF in_buff;
  int buff_status = 0;
  char urlredir[MAX_BUFF];

  int ret, no, found = 0;
  unsigned long int size = 0;
  long double mb;
  const char *virname;
  struct cl_limits limits;

  CURL *eh;
  double usize=0;
  int ft;
  char* p = NULL;
  char err[CURL_ERROR_SIZE];

  /* go into interactive mode if we're run as root */
  if((int)getuid() == 0) {
    interactive = 1;
    fprintf(stderr, "SquidClamav running as UID 0: writing logs to stderr\n");
  }

  /* check for alternate config file given as first argument */
  if(argc == 2)
    alternate_config = argv[1];

  if(argc > 2) {
    fprintf(stderr, "squidclamav:invalid arguments\n");
    fprintf(stderr, "Usage: %s [alternate-squidclamav.conf-file]\n", argv[0]);
     usleep(1000000); /* don't do DoS */
    exit(1);
  }

  /* make stdout line buffered */
  if(setvbuf(stdout, NULL, _IOLBF, 0) != 0) {
     fprintf(stderr, "stdout not line buffered exiting\n");
     usleep(1000000); /* don't do DoS */
     exit(2);
  }

    bridge_mode = 0;
    
    /*********************
      read config files
      into linked lists
      ********************/
    init_ip_list();
    
    init_pattern_list();
    load_patterns();

    signal(SIGHUP,sighup); /* set function calls on signal trap*/
    signal(SIGTERM,sigterm);

    /* work on temporary file */ 
    if ((tf=tmpfile()) == NULL) {
	log(LOG_FILE,"could not open tempfile()\n");
	bridge_mode = 1;
    }
    
    if(bridge_mode)
      log(LOG_FILE, "Invalid condition - continuing in BRIDGE mode\n");
    
    log(LOG_FILE, "SquidClamav (PID %d) started\n", (int)getpid());


	/* load all available databases from default directory */
	if ((ret = cl_loaddbdir(cl_retdbdir(), &root, &no))) {
#ifdef DEBUG
		log(LOG_FILE, "cl_loaddbdir: %s\n", cl_perror(ret));
#endif
		bridge_mode = 1;
	} else {
#ifdef DEBUG
		log(LOG_FILE, "Loaded %d signatures.\n", no);
#endif
		/* build engine */
		if ((ret = cl_build(root))) {
#ifdef DEBUG
			log(LOG_FILE, "Database initialization error: %s\n", cl_strerror(ret));
#endif
			cl_free(root);
			bridge_mode = 1;
		} else {
			/* set up archive limits */
			memset(&limits, 0, sizeof(struct cl_limits));
			limits.maxfiles = 1000; /* max files */
			limits.maxfilesize = 10 * 1048576; /* maximal archived file size == 10 Mb */
			limits.maxreclevel = 5; /* maximal recursion level */
			limits.maxratio = 200; /* maximal compression ratio */
			limits.archivememlim = 0; /* disable memory limit for bzip2 scanner */
		}
	}
    if(bridge_mode) {
      log(LOG_FILE, "Invalid condition - continuing in BRIDGE mode\n");
    } else {

	/* curl init */
	curl_global_init(CURL_GLOBAL_ALL);
	/* get an easy handle */
	if ((eh = curl_easy_init()) == NULL) {
		log(LOG_FILE,"Curl easy initialization failed\n");
		curl_global_cleanup();
		bridge_mode = 1;
		log(LOG_FILE, "Invalid condition - continuing in BRIDGE mode\n");
	} else {
		/* set the curl output to tmeporary file */
		curl_easy_setopt(eh, CURLOPT_WRITEDATA, tf);
		/* set the error buffer */
		curl_easy_setopt(eh, CURLOPT_ERRORBUFFER, err);
	}
    }

    while (fgets(buff, MAX_BUFF, stdin) != NULL) {
      
      /* if configs are completely invalid or some other
	 exception occurs where we want the redirector to
	 continue operation (so that Squid still works!),
	 we simply echo stdin to stdout - i.e. "bridge mode" :-) */
      if(bridge_mode) {
	fprintf(stdout, "\n");
	continue;
      }

      /* separate the four fields
	from the single input line 
	of stdin */

      buff_status = load_in_buff(buff, &in_buff);

      /* if four fields couldn't be separated, or the
	 converted values aren't appropriate, then
	 just echo back the line from stdin */

      if(buff_status == 1) {
	fprintf(stdout, "\n");
	continue;
      }

      /* now we can compare the URL */
      found = pattern_compare(in_buff.url);
      /* -1 mean abort match => No antivir check */
      /* if there's a regex and no content type pattern search and the regex
	 pattern is not found => No antivir check */
      if ((found == -1) || ((check_content_type == 0) && (has_regex == 1) && (found == 0))) {
  
#ifdef DEBUG
	log(LOG_FILE, "No antivir check for url:%s\n", in_buff.url);
#endif
	/* Size exceed limit: workaround to bogus body_max_size */
	if (size_limit > 0) {

		/* go back on top of temporary file */
		rewind(tf);
		ftruncate(fileno(tf), 0); /* remove extrat data from file */
		/* set no proxy */
		curl_easy_setopt(eh, CURLOPT_PROXY, "");
		/* set the url */
		curl_easy_setopt(eh, CURLOPT_URL, in_buff.url);
		/* only get the header to check size and content type */
		curl_easy_setopt(eh, CURLOPT_NOBODY, TRUE);

		/* download the header */
		if (curl_easy_perform(eh) != 0) {
		    /* no header so just go back to squid */
		    log(LOG_FILE,"No header from url %s\n", in_buff.url);
		    fprintf(stdout, "\n");
		    continue;
		}
		/* get the Content Length from the header */
		if (curl_easy_getinfo(eh, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &usize) != CURLE_OK) {
		    /* not ok --> squid should handle this */
		    log(LOG_FILE,"No content length from url %s\n", in_buff.url);
		    fprintf(stdout, "\n");
		    continue;
		}
#ifdef DEBUG
		log(LOG_FILE,"File size is %'.2f, (max size = %d)\n", usize, maxsize_url);
#endif

		if (usize > size_limit) {
			log(LOG_FILE, "Size limit exceed: %'.2f (> %d), Ip source %s %s\n", usize, size_limit, in_buff.src_address, in_buff.ident);
			sprintf(urlredir, "%s?url=%s&size=%'.2f&limit=%d", redirect_url, in_buff.url, usize,size_limit);
			log(LOG_FILE, "Redirecting URL to : %s\n", urlredir);
			strcpy(in_buff.url, urlredir);
			printf("%s %s %s %s\n", in_buff.url, in_buff.src_address, in_buff.ident, in_buff.method);
			continue;
		}
	}
	/* no replacement for the URL was found */
	fprintf(stdout, "\n");
	continue;

      } else {

	/* go back on top of temporary file */
	rewind(tf);
	ftruncate(fileno(tf), 0); /* remove extrat data from file */
	/* set no proxy */
	curl_easy_setopt(eh, CURLOPT_PROXY, "");
        /* set the url */
        curl_easy_setopt(eh, CURLOPT_URL, in_buff.url);
        /* only get the header to check size and content type */
        curl_easy_setopt(eh, CURLOPT_NOBODY, TRUE);

        /* download the header */
        if (curl_easy_perform(eh) != 0) {
            /* no header so just go back to squid */
	    log(LOG_FILE,"No header from url %s\n", in_buff.url);
	    fprintf(stdout, "\n");
            continue;
        }
        /* get the Content Length from the header */
        if (curl_easy_getinfo(eh, CURLINFO_CONTENT_LENGTH_DOWNLOAD, &usize) != CURLE_OK) {
            /* not ok --> squid should handle this */
	    log(LOG_FILE,"No content length from url %s\n", in_buff.url);
	    fprintf(stdout, "\n");
            continue;
        }
#ifdef DEBUG
	log(LOG_FILE,"File size is %'.2f, (max size = %d)\n", usize, maxsize_url);
#endif

        /* get the CONTENT_TYPE out of the header */
	if (check_content_type == 1) {
		if (curl_easy_getinfo(eh, CURLINFO_CONTENT_TYPE, &p) != CURLE_OK) {
		    /* not ok --> squid should handle this */
		    log(LOG_FILE,"No content type from url %s\n", in_buff.url);
		    fprintf(stdout, "\n");
		    continue;
		}
#ifdef DEBUG
		log(LOG_FILE,"Content type is %s\n", p);
#endif
		if ((p == NULL) || (pattern_compare(p) == 0)) {
  
#ifdef DEBUG
		    log(LOG_FILE, "No antivir check for content-type: %s\n", p);
#endif
		    /* no replacement for the URL was found */
		    fprintf(stdout, "\n");
		    continue;
		}
	}
	/* Size exceed limit: workaround to bogus body_max_size */
	if ((size_limit > 0) && (usize > size_limit)) {
		log(LOG_FILE, "Size limit exceed: %'.2f (> %d), Ip source %s %s\n", usize, size_limit, in_buff.src_address, in_buff.ident);
		sprintf(urlredir, "%s?url=%s&size=%'.2f&limit=%d", redirect_url, in_buff.url, usize,size_limit);
		log(LOG_FILE, "Redirecting URL to : %s\n", urlredir);
		strcpy(in_buff.url, urlredir);
		printf("%s %s %s %s\n", in_buff.url, in_buff.src_address, in_buff.ident, in_buff.method);
		continue;
	} else if (usize > maxsize_url) {
            /* file too big --> don't scan */
	    log(LOG_FILE,"File too big (>%d) from url %s, virus scan aborted\n", maxsize_url, in_buff.url);
	    fprintf(stdout, "\n");
            continue;
        } else if (usize == 0) {
#ifdef DEBUG
	    log(LOG_FILE,"Content length of url %s is empty\n", in_buff.url);
#endif
	    fprintf(stdout, "\n");
            continue;
        }

        /* now get the body content */
	rewind(tf);
	ftruncate(fileno(tf), 0); /* remove extrat data from file */
	/* set proxy */
	curl_easy_setopt(eh, CURLOPT_PROXY, MY_PROXY);
	/* download body */
        curl_easy_setopt(eh, CURLOPT_NOBODY, FALSE);
	curl_easy_setopt(eh, CURLOPT_HTTPGET, TRUE);
        if (curl_easy_perform(eh) != 0) {
	    log(LOG_FILE,"Error when downloading url %s\n", in_buff.url);
	    log(LOG_FILE,"CURLOPT_ERRORBUFFER: %s\n", err);
	    fprintf(stdout, "\n");
            continue;
        }
	rewind(tf);
	ft = fileno(tf);
#ifdef DEBUG
	log(LOG_FILE, "Checking %s against ClamAv\n", in_buff.url);
#endif

	/* scan descriptor */
	if((ret = cl_scandesc(ft, &virname, &size, root, &limits, CL_SCAN_STDOPT)) == CL_VIRUS) {
		log(LOG_FILE, "Virus detected: %s, Ip source: %s %s\n", virname, in_buff.src_address, in_buff.ident);
		sprintf(urlredir, "%s?url=%s&virus=%s", redirect_url, in_buff.url, virname);
		log(LOG_FILE, "Redirecting URL to : %s\n", urlredir);
		strcpy(in_buff.url, urlredir);
		fprintf(stdout, "%s %s %s %s\n", in_buff.url, in_buff.src_address, in_buff.ident, in_buff.method);
      } else {
#ifdef DEBUG
		log(LOG_FILE, "No virus detected.\n");
#endif
		if(ret != CL_CLEAN) {
			log(LOG_FILE, "Error: %s\n", cl_perror(ret));
		}
		fprintf(stdout, "\n");
	}

	mb = size * (CL_COUNT_PRECISION / 1024) / 1024.0;
#ifdef DEBUG
	log(LOG_FILE, "Data scanned: %2.2Lf Mb\n", mb);
	log(LOG_FILE, "Returned: %s %s %s %s\n", in_buff.url, in_buff.src_address, in_buff.ident, in_buff.method);
#endif
	continue;
      }
  }
  /* cleanup */
  curl_global_cleanup();
  if (tf != NULL);
	fclose(tf);
  if (root != NULL);
	cl_free(root);
  free(redirect_url);  
  free(alternate_config);  
  free_ip_list();
  free_plist();
  
  return 0;
}

void sighup()
{
  signal(SIGHUP,sighup); /* reset signal */
  log(LOG_FILE, "Sighup received, reloading virus database...\n");
  /* simply terminate the program, squid should reload it */
  /* cleanup */
  curl_global_cleanup();
  if (tf != NULL);
	fclose(tf);
  if (root != NULL)
	cl_free(root);
  free(redirect_url);  
  free(alternate_config);  
  free_ip_list();
  free_plist();
  exit(0);
}

void sigterm()
{
  signal(SIGTERM,sigterm); /* reset signal */
  log(LOG_FILE, "Sigterm received, cleaning all before quit...\n");
  /* cleanup */
  curl_global_cleanup();
  if (tf != NULL);
	fclose(tf);
  if (root != NULL)
	cl_free(root);
  free(redirect_url);  
  free(alternate_config);  
  free_ip_list();
  free_plist();
  exit(0);
}


